" Vim indent file
" Language:    Terraform
" Maintainer:  Gregory Anders
" Upstream:    https://github.com/hashivim/vim-terraform
" Last Change: 2024-09-03

if exists('b:did_indent')
  finish
endif

runtime! indent/hcl.vim
