" Vim :sleep command


sleep	     " sleep for one second
5 sleep	     " sleep for five seconds
sleep 100m   " sleep for 100 milliseconds
100 sleep m  " sleep for 100 milliseconds

sleep!	     " sleep for one second
5 sleep!     " sleep for five seconds
sleep! 100m  " sleep for 100 milliseconds
100 sleep! m " sleep for 100 milliseconds

