vim9script
# Vim9 Ex command containment


fake &
fake ~
fake @
# :*
fake >
fake <
fake #
fake =
fake !

fake Next
fake X

fake abbreviate
fake abclear
fake aboveleft
fake abstract
fake all
fake amenu
fake anoremenu
fake argadd
fake argdedupe
fake argdelete
fake argdo
fake argedit
fake argglobal
fake arglocal
fake args
fake argument
fake ascii
fake augroup Foo
fake augroup END
fake aunmenu
fake autocmd
fake badd
fake ball
fake balt
fake bdelete
fake behave mswin
fake behave xterm
fake belowright
fake bfirst
fake blast
fake bmodified
fake bnext
fake bNext
fake botright
fake bprevious
fake break
fake breakadd
fake breakdel
fake breaklist
fake brewind
fake browse
fake bufdo
fake buffer
fake buffers
fake bunload
fake bwipeout
fake cabbrev
fake cabclear
fake cabove
fake caddbuffer
fake caddexpr
fake caddfile
fake cafter
fake call
fake catch
fake cbefore
fake cbelow
fake cbottom
fake cbuffer
fake cc
fake cclose
fake cd
fake cdo
fake center
fake cexpr
fake cfdo
fake cfile
fake cfirst
fake cgetbuffer
fake cgetexpr
fake cgetfile
fake changes
fake chdir
fake checkpath
fake checktime
fake chistory
fake class
fake clast
fake clearjumps
fake clist
fake close
fake cmap
fake cmapclear
fake cmenu
fake cnewer
fake cnext
fake cNext
fake cnfile
fake cNfile
fake cnoreabbrev
fake cnoremap
fake cnoremenu
fake colder
fake colorscheme
fake comclear
fake command
fake compiler
fake confirm
fake const
fake continue
fake copen
fake copy
fake cpfile
fake cprevious
fake cquit
fake crewind
fake cscope
fake cstag
fake cunabbrev
fake cunmap
fake cunmenu
fake cwindow
fake debug
fake debuggreedy
fake def
fake defcompile
fake defer
fake delcommand
fake delete
fake delfunction
fake delmarks
fake diffget
fake diffoff
fake diffpatch
fake diffput
fake diffsplit
fake diffthis
fake diffupdate
fake digraphs
fake disassemble
fake display
fake djump
fake dlist
fake doautoall
fake doautocmd
fake drop
fake dsearch
fake dsplit
fake earlier
fake echo
fake echoconsole
fake echoerr
fake echohl
fake echomsg
fake echon
fake echowindow
fake edit
fake else
fake elseif
fake emenu
fake endclass
fake enddef
fake endenum
fake endfor
fake endfunction
fake endif
fake endinterface
fake endtry
fake endwhile
fake enew
fake enum
fake eval
fake ex
fake execute
fake exit
fake export
fake exusage
fake file
fake files
fake filetype
fake filter
fake final
fake finally
fake find
fake finish
fake first
fake fixdel
fake fold
fake foldclose
fake folddoclosed
fake folddoopen
fake foldopen
fake for
fake function
fake global
fake goto
fake grep
fake grepadd
fake gui
fake gvim
fake hardcopy
fake help
fake helpclose
fake helpfind
fake helpgrep
fake helptags
fake hide
fake highlight
fake history
fake horizontal
fake iabbrev
fake iabclear
fake if
fake ijump
fake ilist
fake imap
fake imapclear
fake imenu
fake import
fake inoreabbrev
fake inoremap
fake inoremenu
fake interface
fake intro
fake isearch
fake isplit
fake iunabbrev
fake iunmap
fake iunmenu
fake join
fake jumps
fake keepalt
fake keepjumps
fake keepmarks
fake keeppatterns
fake labove
fake laddbuffer
fake laddexpr
fake laddfile
fake lafter
fake language
fake last
fake later
fake lbefore
fake lbelow
fake lbottom
fake lbuffer
fake lcd
fake lchdir
fake lclose
fake lcscope
fake ldo
fake left
fake leftabove
fake legacy
fake lexpr
fake lfdo
fake lfile
fake lfirst
fake lgetbuffer
fake lgetexpr
fake lgetfile
fake lgrep
fake lgrepadd
fake lhelpgrep
fake lhistory
fake list
fake ll
fake llast
fake llist
fake lmake
fake lmap
fake lmapclear
fake lnewer
fake lnext
fake lNext
fake lnfile
fake lNfile
fake lnoremap
fake loadkeymap
fake loadview
fake lockmarks
fake lockvar
fake lolder
fake lopen
fake lpfile
fake lprevious
fake lrewind
fake ls
fake ltag
fake lua
fake luado
fake luafile
fake lunmap
fake lvimgrep
fake lvimgrepadd
fake lwindow
fake make
fake map
fake mapclear
fake mark
fake marks
fake match
fake menu
fake menutranslate
fake messages
fake mkexrc
fake mksession
fake mkspell
fake mkview
fake mkvimrc
fake move
fake mzfile
fake mzscheme
fake nbclose
fake nbkey
fake nbstart
fake new
fake next
fake nmap
fake nmapclear
fake nmenu
fake nnoremap
fake nnoremenu
fake noautocmd
fake nohlsearch
fake noreabbrev
fake noremap
fake noremenu
fake normal
fake noswapfile
fake number
fake nunmap
fake nunmenu
fake oldfiles
fake omap
fake omapclear
fake omenu
fake only
fake onoremap
fake onoremenu
fake options
fake ounmap
fake ounmenu
fake ownsyntax
fake packadd
fake packloadall
fake pclose
fake pedit
fake perl
fake perldo
fake pop
fake popup
fake ppop
fake preserve
fake previous
fake print
fake profdel
fake profile
fake promptfind
fake promptrepl
fake psearch
fake ptag
fake ptfirst
fake ptjump
fake ptlast
fake ptnext
fake ptNext
fake ptprevious
fake ptrewind
fake ptselect
fake put
fake pwd
fake py3
fake py3do
fake py3file
fake pydo
fake pyfile
fake python
fake python3
fake pythonx
fake pyx
fake pyxdo
fake pyxfile
fake qall
fake quit
fake quitall
fake read
fake recover
fake redir
fake redo
fake redraw
fake redrawstatus
fake redrawtabline
fake registers
fake resize
fake retab
fake return
fake rewind
fake right
fake rightbelow
fake ruby
fake rubydo
fake rubyfile
fake rundo
fake runtime
fake rviminfo
fake sall
fake sandbox
fake sargument
fake saveas
fake sball
fake sbfirst
fake sblast
fake sbmodified
fake sbnext
fake sbNext
fake sbprevious
fake sbrewind
fake sbuffer
fake scriptencoding
fake scriptnames
fake scriptversion
fake scscope
fake set
fake setfiletype
fake setglobal
fake setlocal
fake sfind
fake sfirst
fake shell
fake sign
fake silent
fake simalt
fake slast
fake sleep
fake smagic
fake smap
fake smapclear
fake smenu
fake smile
fake snext
fake sNext
fake snomagic
fake snoremap
fake snoremenu
fake sort
fake source
fake spelldump
fake spellgood
fake spellinfo
fake spellrare
fake spellrepall
fake spellundo
fake spellwrong
fake split
fake sprevious
fake srewind
fake stag
fake startgreplace
fake startinsert
fake startreplace
fake stjump
fake stop
fake stopinsert
fake stselect
fake substitute
fake sunhide
fake sunmap
fake sunmenu
fake suspend
fake sview
fake swapname
fake syncbind
fake syntax
fake syntime
fake tab
fake tabclose
fake tabdo
fake tabedit
fake tabfind
fake tabfirst
fake tablast
fake tabmove
fake tabnew
fake tabnext
fake tabNext
fake tabonly
fake tabprevious
fake tabrewind
fake tabs
fake tag
fake tags
fake tcd
fake tchdir
fake tcl
fake tcldo
fake tclfile
fake tearoff
fake terminal
fake tfirst
fake throw
fake tjump
fake tlast
fake tlmenu
fake tlnoremenu
fake tlunmenu
fake tmap
fake tmapclear
fake tmenu
fake tnext
fake tNext
fake tnoremap
fake topleft
fake tprevious
fake trewind
fake try
fake tselect
fake tunmap
fake tunmenu
fake type
fake unabbreviate
fake undo
fake undojoin
fake undolist
fake unhide
fake uniq
fake unlet
fake unlockvar
fake unmap
fake unmenu
fake unsilent
fake update
fake var
fake verbose
fake version
fake vertical
fake vglobal
fake view
fake vim9cmd
# :vim9script
fake vimgrep
fake vimgrepadd
fake visual
fake viusage
fake vmap
fake vmapclear
fake vmenu
fake vnew
fake vnoremap
fake vnoremenu
fake vsplit
fake vunmap
fake vunmenu
fake wall
fake while
fake wincmd
fake windo
fake winpos
fake winsize
fake wnext
fake wNext
fake wprevious
fake wq
fake wqall
fake write
fake wundo
fake wviminfo
fake xall
fake xmap
fake xmapclear
fake xmenu
fake xnoremap
fake xnoremenu
fake xrestore
fake xunmap
fake xunmenu
fake yank
fake z


# Legacy-script only

fake Print
fake append
fake change
fake insert
fake k
fake let
fake mode
fake open
fake t
fake xit


# User commands

fake Foobar
fake Foobar!
fake Foobar123
fake Foobar123!
fake Foo123bar
fake Foo123bar!

