" Vim filetype plugin
" Language:	JSON-LD
" Maintainer:	The Vim Project <https://github.com/vim/vim>
" Last Change:	2026 Aug 06

if exists('b:did_ftplugin')
  finish
endif

runtime! ftplugin/json.vim
