vim9script
# Vim9 Ex commands


:help
 :help
: help
 : help

:&
:~
:@
# :*
:>
:<
:#
:=
:!

:Next
:X

:abbreviate
:abclear
:aboveleft
:abstract
:all
:amenu
:anoremenu
:argadd
:argdedupe
:argdelete
:argdo
:argedit
:argglobal
:arglocal
:args
:argument
:ascii
:augroup Foo
:augroup END
:aunmenu
:autocmd
:badd
:ball
:balt
:bdelete
:behave mswin
:behave xterm
:belowright
:bfirst
:blast
:bmodified
:bnext
:bNext
:botright
:bprevious
:break
:breakadd
:breakdel
:breaklist
:brewind
:browse
:bufdo
:buffer
:buffers
:bunload
:bwipeout
:cabbrev
:cabclear
:cabove
:caddbuffer
:caddexpr
:caddfile
:cafter
:call
:catch
:cbefore
:cbelow
:cbottom
:cbuffer
:cc
:cclose
:cd
:cdo
:center
:cexpr
:cfdo
:cfile
:cfirst
:cgetbuffer
:cgetexpr
:cgetfile
:changes
:chdir
:checkpath
:checktime
:chistory
:class
endclass
:clast
:clearjumps
:clist
:close
:cmap
:cmapclear
:cmenu
:cnewer
:cnext
:cNext
:cnfile
:cNfile
:cnoreabbrev
:cnoremap
:cnoremenu
:colder
:colorscheme
:comclear
:command
:compiler
:confirm
:const
:continue
:copen
:copy
:cpfile
:cprevious
:cquit
:crewind
:cscope
:cstag
:cunabbrev
:cunmap
:cunmenu
:cwindow
:debug
:debuggreedy
:def
:defcompile
:defer
:delcommand
:delete
:delfunction
:delmarks
:diffget
:diffoff
:diffpatch
:diffput
:diffsplit
:diffthis
:diffupdate
:digraphs
:disassemble
:display
:djump
:dlist
:doautoall
:doautocmd
:drop
:dsearch
:dsplit
:earlier
:echo
:echoconsole
:echoerr
:echohl
:echomsg
:echon
:echowindow
:edit
:else
:elseif
:emenu
# :endclass
:enddef
# :endenum
:endfor
:endfunction
:endif
# :endinterface
:endtry
:endwhile
:enew
:enum
endenum
:eval
:ex
:execute
:exit
:export
:exusage
:file
:files
:filetype
:filter
:final
:finally
:find
:finish
:first
:fixdel
:fold
:foldclose
:folddoclosed
:folddoopen
:foldopen
:for foo in bar | endfor
:function
:global
:goto
:grep
:grepadd
:gui
:gvim
:hardcopy
:help
:helpclose
:helpfind
:helpgrep
:helptags
:hide
:highlight
:history
:horizontal
:iabbrev
:iabclear
:if
:ijump
:ilist
:imap
:imapclear
:imenu
:import
:inoreabbrev
:inoremap
:inoremenu
:interface
endinterface
:intro
:isearch
:isplit
:iunabbrev
:iunmap
:iunmenu
:join
:jumps
:keepalt
:keepjumps
:keepmarks
:keeppatterns
:labove
:laddbuffer
:laddexpr
:laddfile
:lafter
:language
:last
:later
:lbefore
:lbelow
:lbottom
:lbuffer
:lcd
:lchdir
:lclose
:lcscope
:ldo
:left
:leftabove
:legacy
:lexpr
:lfdo
:lfile
:lfirst
:lgetbuffer
:lgetexpr
:lgetfile
:lgrep
:lgrepadd
:lhelpgrep
:lhistory
:list
:ll
:llast
:llist
:lmake
:lmap
:lmapclear
:lnewer
:lnext
:lNext
:lnfile
:lNfile
:lnoremap
# :loadkeymap # disabled - runs until EOF
:loadview
:lockmarks
:lockvar
:lolder
:lopen
:lpfile
:lprevious
:lrewind
:ls
:ltag
:lua
:luado
:luafile
:lunmap
:lvimgrep
:lvimgrepadd
:lwindow
:make
:map
:mapclear
:mark
:marks
:match
:menu
:menutranslate
:messages
:mkexrc
:mksession
:mkspell
:mkview
:mkvimrc
:move
:mzfile
:mzscheme
:nbclose
:nbkey
:nbstart
:new
:next
:nmap
:nmapclear
:nmenu
:nnoremap
:nnoremenu
:noautocmd
:nohlsearch
:noreabbrev
:noremap
:noremenu
:normal
:noswapfile
:number
:nunmap
:nunmenu
:oldfiles
:omap
:omapclear
:omenu
:only
:onoremap
:onoremenu
:options
:ounmap
:ounmenu
:ownsyntax
:packadd
:packloadall
:pclose
:pedit
:perl
:perldo
:pop
:popup
:ppop
:preserve
:previous
:print
:profdel
:profile
:promptfind
:promptrepl
:psearch
:ptag
:ptfirst
:ptjump
:ptlast
:ptnext
:ptNext
:ptprevious
:ptrewind
:ptselect
:put
:pwd
:py3
:py3do
:py3file
:pydo
:pyfile
:python
:python3
:pythonx
:pyx
:pyxdo
:pyxfile
:qall
:quit
:quitall
:read
:recover
:redir
:redo
:redraw
:redrawstatus
:redrawtabline
:registers
:resize
:retab
:return
:rewind
:right
:rightbelow
:ruby
:rubydo
:rubyfile
:rundo
:runtime
:rviminfo
:sall
:sandbox
:sargument
:saveas
:sball
:sbfirst
:sblast
:sbmodified
:sbnext
:sbNext
:sbprevious
:sbrewind
:sbuffer
:scriptencoding
:scriptnames
:scriptversion
:scscope
:set
:setfiletype
:setglobal
:setlocal
:sfind
:sfirst
:shell
:sign
:silent
:simalt
:slast
:sleep
:smagic
:smap
:smapclear
:smenu
:smile
:snext
:sNext
:snomagic
:snoremap
:snoremenu
:sort
:source
:spelldump
:spellgood
:spellinfo
:spellrare
:spellrepall
:spellundo
:spellwrong
:split
:sprevious
:srewind
:stag
:startgreplace
:startinsert
:startreplace
:stjump
:stop
:stopinsert
:stselect
:substitute
:sunhide
:sunmap
:sunmenu
:suspend
:sview
:swapname
:syncbind
:syntax
:syntime
:tab
:tabclose
:tabdo
:tabedit
:tabfind
:tabfirst
:tablast
:tabmove
:tabnew
:tabnext
:tabNext
:tabonly
:tabprevious
:tabrewind
:tabs
:tag
:tags
:tcd
:tchdir
:tcl
:tcldo
:tclfile
:tearoff
:terminal
:tfirst
:throw
:tjump
:tlast
:tlmenu
:tlnoremenu
:tlunmenu
:tmap
:tmapclear
:tmenu
:tnext
:tNext
:tnoremap
:topleft
:tprevious
:trewind
:try
:tselect
:tunmap
:tunmenu
:type
:unabbreviate
:undo
:undojoin
:undolist
:unhide
:uniq
:unlet
:unlockvar
:unmap
:unmenu
:unsilent
:update
:var
:verbose
:version
:vertical
:vglobal
:view
:vim9cmd
# :vim9script
:vimgrep
:vimgrepadd
:visual
:viusage
:vmap
:vmapclear
:vmenu
:vnew
:vnoremap
:vnoremenu
:vsplit
:vunmap
:vunmenu
:wall
:while
:wincmd
:windo
:winpos
:winsize
:wnext
:wNext
:wprevious
:wq
:wqall
:write
:wundo
:wviminfo
:xall
:xmap
:xmapclear
:xmenu
:xnoremap
:xnoremenu
:xrestore
:xunmap
:xunmenu
:yank
:z

Foo()|help
Foo() |help
Foo()| help
Foo() | help

Foo() |:help
Foo() | :help
Foo() |: help
Foo() | : help

Foo() | &
Foo() | ~
Foo() | @
# Foo() | *
Foo() | >
Foo() | <
Foo() | #
Foo() | =
Foo() | !


Foo() | Next
Foo() | X

Foo() | abbreviate
Foo() | abclear
Foo() | aboveleft
Foo() | abstract
Foo() | all
Foo() | amenu
Foo() | anoremenu
Foo() | argadd
Foo() | argdedupe
Foo() | argdelete
Foo() | argdo
Foo() | argedit
Foo() | argglobal
Foo() | arglocal
Foo() | args
Foo() | argument
Foo() | ascii
Foo() | augroup Foo
Foo() | augroup END
Foo() | aunmenu
Foo() | autocmd
Foo() | badd
Foo() | ball
Foo() | balt
Foo() | bdelete
Foo() | behave mswin
Foo() | behave xterm
Foo() | belowright
Foo() | bfirst
Foo() | blast
Foo() | bmodified
Foo() | bnext
Foo() | bNext
Foo() | botright
Foo() | bprevious
Foo() | break
Foo() | breakadd
Foo() | breakdel
Foo() | breaklist
Foo() | brewind
Foo() | browse
Foo() | bufdo
Foo() | buffer
Foo() | buffers
Foo() | bunload
Foo() | bwipeout
Foo() | cabbrev
Foo() | cabclear
Foo() | cabove
Foo() | caddbuffer
Foo() | caddexpr
Foo() | caddfile
Foo() | cafter
Foo() | call
Foo() | catch
Foo() | cbefore
Foo() | cbelow
Foo() | cbottom
Foo() | cbuffer
Foo() | cc
Foo() | cclose
Foo() | cd
Foo() | cdo
Foo() | center
Foo() | cexpr
Foo() | cfdo
Foo() | cfile
Foo() | cfirst
Foo() | cgetbuffer
Foo() | cgetexpr
Foo() | cgetfile
Foo() | changes
Foo() | chdir
Foo() | checkpath
Foo() | checktime
Foo() | chistory
Foo() | class
endclass
Foo() | clast
Foo() | clearjumps
Foo() | clist
Foo() | close
Foo() | cmap
Foo() | cmapclear
Foo() | cmenu
Foo() | cnewer
Foo() | cnext
Foo() | cNext
Foo() | cnfile
Foo() | cNfile
Foo() | cnoreabbrev
Foo() | cnoremap
Foo() | cnoremenu
Foo() | colder
Foo() | colorscheme
Foo() | comclear
Foo() | command
Foo() | compiler
Foo() | confirm
Foo() | const
Foo() | continue
Foo() | copen
Foo() | copy
Foo() | cpfile
Foo() | cprevious
Foo() | cquit
Foo() | crewind
Foo() | cscope
Foo() | cstag
Foo() | cunabbrev
Foo() | cunmap
Foo() | cunmenu
Foo() | cwindow
Foo() | debug
Foo() | debuggreedy
Foo() | def
Foo() | defcompile
Foo() | defer
Foo() | delcommand
Foo() | delete
Foo() | delfunction
Foo() | delmarks
Foo() | diffget
Foo() | diffoff
Foo() | diffpatch
Foo() | diffput
Foo() | diffsplit
Foo() | diffthis
Foo() | diffupdate
Foo() | digraphs
Foo() | disassemble
Foo() | display
Foo() | djump
Foo() | dlist
Foo() | doautoall
Foo() | doautocmd
Foo() | drop
Foo() | dsearch
Foo() | dsplit
Foo() | earlier
Foo() | echo
Foo() | echoconsole
Foo() | echoerr
Foo() | echohl
Foo() | echomsg
Foo() | echon
Foo() | echowindow
Foo() | edit
Foo() | else
Foo() | elseif
Foo() | emenu
# :endclass
Foo() | enddef
# :endenum
Foo() | endfor
Foo() | endfunction
Foo() | endif
# :endinterface
Foo() | endtry
Foo() | endwhile
Foo() | enew
Foo() | enum
endenum
Foo() | eval
Foo() | ex
Foo() | execute
Foo() | exit
Foo() | export
Foo() | exusage
Foo() | file
Foo() | files
Foo() | filetype
Foo() | filter
Foo() | final
Foo() | finally
Foo() | find
Foo() | finish
Foo() | first
Foo() | fixdel
Foo() | fold
Foo() | foldclose
Foo() | folddoclosed
Foo() | folddoopen
Foo() | foldopen
Foo() | for foo in bar | endfor
Foo() | function
Foo() | global
Foo() | goto
Foo() | grep
Foo() | grepadd
Foo() | gui
Foo() | gvim
Foo() | hardcopy
Foo() | help
Foo() | helpclose
Foo() | helpfind
Foo() | helpgrep
Foo() | helptags
Foo() | hide
Foo() | highlight
Foo() | history
Foo() | horizontal
Foo() | iabbrev
Foo() | iabclear
Foo() | if
Foo() | ijump
Foo() | ilist
Foo() | imap
Foo() | imapclear
Foo() | imenu
Foo() | import
Foo() | inoreabbrev
Foo() | inoremap
Foo() | inoremenu
Foo() | interface
endinterface
Foo() | intro
Foo() | isearch
Foo() | isplit
Foo() | iunabbrev
Foo() | iunmap
Foo() | iunmenu
Foo() | join
Foo() | jumps
Foo() | keepalt
Foo() | keepjumps
Foo() | keepmarks
Foo() | keeppatterns
Foo() | labove
Foo() | laddbuffer
Foo() | laddexpr
Foo() | laddfile
Foo() | lafter
Foo() | language
Foo() | last
Foo() | later
Foo() | lbefore
Foo() | lbelow
Foo() | lbottom
Foo() | lbuffer
Foo() | lcd
Foo() | lchdir
Foo() | lclose
Foo() | lcscope
Foo() | ldo
Foo() | left
Foo() | leftabove
Foo() | legacy
Foo() | lexpr
Foo() | lfdo
Foo() | lfile
Foo() | lfirst
Foo() | lgetbuffer
Foo() | lgetexpr
Foo() | lgetfile
Foo() | lgrep
Foo() | lgrepadd
Foo() | lhelpgrep
Foo() | lhistory
Foo() | list
Foo() | ll
Foo() | llast
Foo() | llist
Foo() | lmake
Foo() | lmap
Foo() | lmapclear
Foo() | lnewer
Foo() | lnext
Foo() | lNext
Foo() | lnfile
Foo() | lNfile
Foo() | lnoremap
# :loadkeymap # disabled - runs until EOF
Foo() | loadview
Foo() | lockmarks
Foo() | lockvar
Foo() | lolder
Foo() | lopen
Foo() | lpfile
Foo() | lprevious
Foo() | lrewind
Foo() | ls
Foo() | ltag
Foo() | lua
Foo() | luado
Foo() | luafile
Foo() | lunmap
Foo() | lvimgrep
Foo() | lvimgrepadd
Foo() | lwindow
Foo() | make
Foo() | map
Foo() | mapclear
Foo() | mark
Foo() | marks
Foo() | match
Foo() | menu
Foo() | menutranslate
Foo() | messages
Foo() | mkexrc
Foo() | mksession
Foo() | mkspell
Foo() | mkview
Foo() | mkvimrc
Foo() | move
Foo() | mzfile
Foo() | mzscheme
Foo() | nbclose
Foo() | nbkey
Foo() | nbstart
Foo() | new
Foo() | next
Foo() | nmap
Foo() | nmapclear
Foo() | nmenu
Foo() | nnoremap
Foo() | nnoremenu
Foo() | noautocmd
Foo() | nohlsearch
Foo() | noreabbrev
Foo() | noremap
Foo() | noremenu
Foo() | normal
Foo() | noswapfile
Foo() | number
Foo() | nunmap
Foo() | nunmenu
Foo() | oldfiles
Foo() | omap
Foo() | omapclear
Foo() | omenu
Foo() | only
Foo() | onoremap
Foo() | onoremenu
Foo() | options
Foo() | ounmap
Foo() | ounmenu
Foo() | ownsyntax
Foo() | packadd
Foo() | packloadall
Foo() | pclose
Foo() | pedit
Foo() | perl
Foo() | perldo
Foo() | pop
Foo() | popup
Foo() | ppop
Foo() | preserve
Foo() | previous
Foo() | print
Foo() | profdel
Foo() | profile
Foo() | promptfind
Foo() | promptrepl
Foo() | psearch
Foo() | ptag
Foo() | ptfirst
Foo() | ptjump
Foo() | ptlast
Foo() | ptnext
Foo() | ptNext
Foo() | ptprevious
Foo() | ptrewind
Foo() | ptselect
Foo() | put
Foo() | pwd
Foo() | py3
Foo() | py3do
Foo() | py3file
Foo() | pydo
Foo() | pyfile
Foo() | python
Foo() | python3
Foo() | pythonx
Foo() | pyx
Foo() | pyxdo
Foo() | pyxfile
Foo() | qall
Foo() | quit
Foo() | quitall
Foo() | read
Foo() | recover
Foo() | redir
Foo() | redo
Foo() | redraw
Foo() | redrawstatus
Foo() | redrawtabline
Foo() | registers
Foo() | resize
Foo() | retab
Foo() | return
Foo() | rewind
Foo() | right
Foo() | rightbelow
Foo() | ruby
Foo() | rubydo
Foo() | rubyfile
Foo() | rundo
Foo() | runtime
Foo() | rviminfo
Foo() | sall
Foo() | sandbox
Foo() | sargument
Foo() | saveas
Foo() | sball
Foo() | sbfirst
Foo() | sblast
Foo() | sbmodified
Foo() | sbnext
Foo() | sbNext
Foo() | sbprevious
Foo() | sbrewind
Foo() | sbuffer
Foo() | scriptencoding
Foo() | scriptnames
Foo() | scriptversion
Foo() | scscope
Foo() | set
Foo() | setfiletype
Foo() | setglobal
Foo() | setlocal
Foo() | sfind
Foo() | sfirst
Foo() | shell
Foo() | sign
Foo() | silent
Foo() | simalt
Foo() | slast
Foo() | sleep
Foo() | smagic
Foo() | smap
Foo() | smapclear
Foo() | smenu
Foo() | smile
Foo() | snext
Foo() | sNext
Foo() | snomagic
Foo() | snoremap
Foo() | snoremenu
Foo() | sort
Foo() | source
Foo() | spelldump
Foo() | spellgood
Foo() | spellinfo
Foo() | spellrare
Foo() | spellrepall
Foo() | spellundo
Foo() | spellwrong
Foo() | split
Foo() | sprevious
Foo() | srewind
Foo() | stag
Foo() | startgreplace
Foo() | startinsert
Foo() | startreplace
Foo() | stjump
Foo() | stop
Foo() | stopinsert
Foo() | stselect
Foo() | substitute
Foo() | sunhide
Foo() | sunmap
Foo() | sunmenu
Foo() | suspend
Foo() | sview
Foo() | swapname
Foo() | syncbind
Foo() | syntax
Foo() | syntime
Foo() | tab
Foo() | tabclose
Foo() | tabdo
Foo() | tabedit
Foo() | tabfind
Foo() | tabfirst
Foo() | tablast
Foo() | tabmove
Foo() | tabnew
Foo() | tabnext
Foo() | tabNext
Foo() | tabonly
Foo() | tabprevious
Foo() | tabrewind
Foo() | tabs
Foo() | tag
Foo() | tags
Foo() | tcd
Foo() | tchdir
Foo() | tcl
Foo() | tcldo
Foo() | tclfile
Foo() | tearoff
Foo() | terminal
Foo() | tfirst
Foo() | throw
Foo() | tjump
Foo() | tlast
Foo() | tlmenu
Foo() | tlnoremenu
Foo() | tlunmenu
Foo() | tmap
Foo() | tmapclear
Foo() | tmenu
Foo() | tnext
Foo() | tNext
Foo() | tnoremap
Foo() | topleft
Foo() | tprevious
Foo() | trewind
Foo() | try
Foo() | tselect
Foo() | tunmap
Foo() | tunmenu
Foo() | type
Foo() | unabbreviate
Foo() | undo
Foo() | undojoin
Foo() | undolist
Foo() | unhide
Foo() | uniq
Foo() | unlet
Foo() | unlockvar
Foo() | unmap
Foo() | unmenu
Foo() | unsilent
Foo() | update
Foo() | var
Foo() | verbose
Foo() | version
Foo() | vertical
Foo() | vglobal
Foo() | view
Foo() | vim9cmd
# :vim9script
Foo() | vimgrep
Foo() | vimgrepadd
Foo() | visual
Foo() | viusage
Foo() | vmap
Foo() | vmapclear
Foo() | vmenu
Foo() | vnew
Foo() | vnoremap
Foo() | vnoremenu
Foo() | vsplit
Foo() | vunmap
Foo() | vunmenu
Foo() | wall
Foo() | while
Foo() | wincmd
Foo() | windo
Foo() | winpos
Foo() | winsize
Foo() | wnext
Foo() | wNext
Foo() | wprevious
Foo() | wq
Foo() | wqall
Foo() | write
Foo() | wundo
Foo() | wviminfo
Foo() | xall
Foo() | xmap
Foo() | xmapclear
Foo() | xmenu
Foo() | xnoremap
Foo() | xnoremenu
Foo() | xrestore
Foo() | xunmap
Foo() | xunmenu
Foo() | yank

# Legacy-script only

:Print
:append
    text
.
:change
    text
.
:insert
    text
.
:k
:let
:mode
:open
:t
:xit

call Foo() | append
    text
.
call Foo() | change
    text
.
call Foo() | insert
    text
.
call Foo() | k
call Foo() | let
call Foo() | mode
call Foo() | open
call Foo() | t
call Foo() | xit


# User commands

Foobar
Foobar!
Foobar123
Foobar123!
Foo123bar
Foo123bar!

:Foobar
:Foobar!
:Foobar123
:Foobar123!
:Foo123bar
:Foo123bar!

Foo() | Foobar
Foo() | Foobar!
Foo() | Foobar123
Foo() | Foobar123!
Foo() | Foo123bar
Foo() | Foo123bar!

Foo() | :Foobar
Foo() | :Foobar!
Foo() | :Foobar123
Foo() | :Foobar123!
Foo() | :Foo123bar
Foo() | :Foo123bar!

